@@ -19,18 +19,25 @@ import logging
 import threading
 import time
 import warnings
+from typing import Any
+from typing import Dict
+from typing import List
+from typing import Union
 
 import pandas as pd
 
 import apache_beam as beam
 from apache_beam.dataframe.frame_base import DeferredBase
+from apache_beam.options import pipeline_options
 from apache_beam.portability.api import beam_runner_api_pb2
+from apache_beam.runners import runner
 from apache_beam.runners.interactive import background_caching_job as bcj
 from apache_beam.runners.interactive import interactive_environment as ie
 from apache_beam.runners.interactive import interactive_runner as ir
 from apache_beam.runners.interactive import pipeline_fragment as pf
 from apache_beam.runners.interactive import utils
 from apache_beam.runners.interactive.caching.cacheable import CacheKey
+from apache_beam.runners.interactive.options import capture_control
 from apache_beam.runners.runner import PipelineState
 
 _LOGGER = logging.getLogger(__name__)
@@ -40,12 +47,11 @@ class ElementStream:
   """A stream of elements from a given PCollection."""
   def __init__(
       self,
-      pcoll,  # type: beam.pvalue.PCollection
-      var,  # type: str
-      cache_key,  # type: str
-      max_n,  # type: int
-      max_duration_secs  # type: float
-      ):
+      pcoll: beam.pvalue.PCollection,
+      var: str,
+      cache_key: str,
+      max_n: int,
+      max_duration_secs: float):
     self._pcoll = pcoll
     self._cache_key = cache_key
     self._pipeline = ie.current_env().user_pipeline(pcoll.pipeline)
@@ -58,47 +64,37 @@ class ElementStream:
     self._done = False
 
   @property
-  def var(self):
-    # type: () -> str
-
+  def var(self) -> str:
     """Returns the variable named that defined this PCollection."""
     return self._var
 
   @property
-  def pcoll(self):
-    # type: () -> beam.pvalue.PCollection
-
+  def pcoll(self) -> beam.pvalue.PCollection:
     """Returns the PCollection that supplies this stream with data."""
     return self._pcoll
 
   @property
-  def cache_key(self):
-    # type: () -> str
-
+  def cache_key(self) -> str:
     """Returns the cache key for this stream."""
     return self._cache_key
 
-  def display_id(self, suffix):
-    # type: (str) -> str
-
+  def display_id(self, suffix: str) -> str:
     """Returns a unique id able to be displayed in a web browser."""
     return utils.obfuscate(self._cache_key, suffix)
 
-  def is_computed(self):
-    # type: () -> boolean # noqa: F821
+  def is_computed(self) -> bool:
+    # noqa: F821
 
     """Returns True if no more elements will be recorded."""
     return self._pcoll in ie.current_env().computed_pcollections
 
-  def is_done(self):
-    # type: () -> boolean # noqa: F821
+  def is_done(self) -> bool:
+    # noqa: F821
 
     """Returns True if no more new elements will be yielded."""
     return self._done
 
-  def read(self, tail=True):
-    # type: (boolean) -> Any # noqa: F821
-
+  def read(self, tail: bool = True) -> Any:
     """Reads the elements currently recorded."""
 
     # Get the cache manager and wait until the file exists.
@@ -154,19 +150,20 @@ class Recording:
   """A group of PCollections from a given pipeline run."""
   def __init__(
       self,
-      user_pipeline,  # type: beam.Pipeline
-      pcolls,  # type: List[beam.pvalue.PCollection] # noqa: F821
-      result,  # type: beam.runner.PipelineResult
-      max_n,  # type: int
-      max_duration_secs,  # type: float
-      ):
+      user_pipeline: beam.Pipeline,
+      pcolls: List[beam.pvalue.PCollection],  # noqa: F821
+      result: 'beam.runner.PipelineResult',
+      max_n: int,
+      max_duration_secs: float,
+  ):
     self._user_pipeline = user_pipeline
     self._result = result
     self._result_lock = threading.Lock()
     self._pcolls = pcolls
-    pcoll_var = lambda pcoll: {v: k
-                               for k, v in utils.pcoll_by_name().items()}.get(
-                                   pcoll, None)
+    pcoll_var = lambda pcoll: {
+        v: k
+        for k, v in utils.pcoll_by_name().items()
+    }.get(pcoll, None)
 
     self._streams = {
         pcoll: ElementStream(
@@ -188,9 +185,7 @@ class Recording:
     self._mark_computed.daemon = True
     self._mark_computed.start()
 
-  def _mark_all_computed(self):
-    # type: () -> None
-
+  def _mark_all_computed(self) -> None:
     """Marks all the PCollections upon a successful pipeline run."""
     if not self._result:
       return
@@ -216,40 +211,28 @@ class Recording:
     if self._result.state is PipelineState.DONE and self._set_computed:
       ie.current_env().mark_pcollection_computed(self._pcolls)
 
-  def is_computed(self):
-    # type: () -> boolean # noqa: F821
-
+  def is_computed(self) -> bool:
     """Returns True if all PCollections are computed."""
     return all(s.is_computed() for s in self._streams.values())
 
-  def stream(self, pcoll):
-    # type: (beam.pvalue.PCollection) -> ElementStream
-
+  def stream(self, pcoll: beam.pvalue.PCollection) -> ElementStream:
     """Returns an ElementStream for a given PCollection."""
     return self._streams[pcoll]
 
-  def computed(self):
-    # type: () -> None
-
+  def computed(self) -> None:
     """Returns all computed ElementStreams."""
     return {p: s for p, s in self._streams.items() if s.is_computed()}
 
-  def uncomputed(self):
-    # type: () -> None
-
+  def uncomputed(self) -> None:
     """Returns all uncomputed ElementStreams."""
     return {p: s for p, s in self._streams.items() if not s.is_computed()}
 
-  def cancel(self):
-    # type: () -> None
-
+  def cancel(self) -> None:
     """Cancels the recording."""
     with self._result_lock:
       self._result.cancel()
 
-  def wait_until_finish(self):
-    # type: () -> None
-
+  def wait_until_finish(self) -> None:
     """Waits until the pipeline is done and returns the final state.
 
     This also marks any PCollections as computed right away if the pipeline is
@@ -261,9 +244,7 @@ class Recording:
     self._mark_computed.join()
     return self._result.state
 
-  def describe(self):
-    # type: () -> dict[str, int]
-
+  def describe(self) -> Dict[str, int]:
     """Returns a dictionary describing the cache and recording."""
     cache_manager = ie.current_env().get_cache_manager(self._user_pipeline)
 
@@ -274,18 +255,19 @@ class Recording:
 
 class RecordingManager:
   """Manages recordings of PCollections for a given pipeline."""
-  def __init__(self, user_pipeline, pipeline_var=None, test_limiters=None):
-    # type: (beam.Pipeline, str, list[Limiter]) -> None # noqa: F821
+  def __init__(
+      self,
+      user_pipeline: beam.Pipeline,
+      pipeline_var: str = None,
+      test_limiters: List['Limiter'] = None) -> None:  # noqa: F821
 
-    self.user_pipeline = user_pipeline  # type: beam.Pipeline
-    self.pipeline_var = pipeline_var if pipeline_var else ''  # type: str
-    self._recordings = set()  # type: set[Recording]
-    self._start_time_sec = 0  # type: float
+    self.user_pipeline: beam.Pipeline = user_pipeline
+    self.pipeline_var: str = pipeline_var if pipeline_var else ''
+    self._recordings: set[Recording] = set()
+    self._start_time_sec: float = 0
     self._test_limiters = test_limiters if test_limiters else []
 
-  def _watch(self, pcolls):
-    # type: (List[beam.pvalue.PCollection]) -> None # noqa: F821
-
+  def _watch(self, pcolls: List[beam.pvalue.PCollection]) -> None:
     """Watch any pcollections not being watched.
 
     This allows for the underlying caching layer to identify the PCollection as
@@ -314,9 +296,7 @@ class RecordingManager:
         ie.current_env().watch(
             {'anonymous_pcollection_{}'.format(id(pcoll)): pcoll})
 
-  def _clear(self):
-    # type: () -> None
-
+  def _clear(self) -> None:
     """Clears the recording of all non-source PCollections."""
 
     cache_manager = ie.current_env().get_cache_manager(self.user_pipeline)
@@ -338,17 +318,13 @@ class RecordingManager:
     for pc in pcolls:
       cache_manager.clear('full', pc)
 
-  def clear(self):
-    # type: () -> None
-
+  def clear(self) -> None:
     """Clears all cached PCollections for this RecordingManager."""
     cache_manager = ie.current_env().get_cache_manager(self.user_pipeline)
     if cache_manager:
       cache_manager.cleanup()
 
-  def cancel(self):
-    # type: (None) -> None
-
+  def cancel(self: None) -> None:
     """Cancels the current background recording job."""
 
     bcj.attempt_to_cancel_background_caching_job(self.user_pipeline)
@@ -361,9 +337,7 @@ class RecordingManager:
     # evict the BCJ after they complete.
     ie.current_env().evict_background_caching_job(self.user_pipeline)
 
-  def describe(self):
-    # type: () -> dict[str, int]
-
+  def describe(self) -> Dict[str, int]:
     """Returns a dictionary describing the cache and recording."""
 
     cache_manager = ie.current_env().get_cache_manager(self.user_pipeline)
@@ -384,14 +358,14 @@ class RecordingManager:
         'pipeline_var': self.pipeline_var
     }
 
-  def record_pipeline(self):
-    # type: () -> bool
-
+  def record_pipeline(self) -> bool:
     """Starts a background caching job for this RecordingManager's pipeline."""
 
     runner = self.user_pipeline.runner
     if isinstance(runner, ir.InteractiveRunner):
       runner = runner._underlying_runner
+    if hasattr(runner, 'is_interactive'):
+      runner.is_interactive()
 
     # Make sure that sources without a user reference are still cached.
     ie.current_env().add_user_pipeline(self.user_pipeline)
@@ -412,11 +386,24 @@ class RecordingManager:
       return True
     return False
 
-  def record(self, pcolls, max_n, max_duration):
-    # type: (List[beam.pvalue.PCollection], int, Union[int,str]) -> Recording # noqa: F821
+  def record(
+      self,
+      pcolls: List[beam.pvalue.PCollection],
+      *,
+      max_n: int,
+      max_duration: Union[int, str],
+      runner: runner.PipelineRunner = None,
+      options: pipeline_options.PipelineOptions = None,
+      force_compute: bool = False) -> Recording:
+    # noqa: F821
 
     """Records the given PCollections."""
 
+    if not ie.current_env().options.enable_recording_replay:
+      capture_control.evict_captured_data()
+    if force_compute:
+      ie.current_env().evict_computed_pcollections()
+
     # Assert that all PCollection come from the same user_pipeline.
     for pcoll in pcolls:
       assert pcoll.pipeline is self.user_pipeline, (
@@ -448,12 +435,20 @@ class RecordingManager:
       # incomplete.
       self._clear()
 
+      merged_options = pipeline_options.PipelineOptions(
+          **{
+              **self.user_pipeline.options.get_all_options(
+                  drop_default=True, retain_unknown_options=True),
+              **options.get_all_options(
+                  drop_default=True, retain_unknown_options=True)
+          }) if options else self.user_pipeline.options
+
       cache_path = ie.current_env().options.cache_root
       is_remote_run = cache_path and ie.current_env(
       ).options.cache_root.startswith('gs://')
       pf.PipelineFragment(
-          list(uncomputed_pcolls),
-          self.user_pipeline.options).run(blocking=is_remote_run)
+          list(uncomputed_pcolls), merged_options,
+          runner=runner).run(blocking=is_remote_run)
       result = ie.current_env().pipeline_result(self.user_pipeline)
     else:
       result = None
@@ -464,8 +459,13 @@ class RecordingManager:
 
     return recording
 
-  def read(self, pcoll_name, pcoll, max_n, max_duration_secs):
-    # type: (str, beam.pvalue.PValue, int, float) -> Union[None, ElementStream] # noqa: F821
+  def read(
+      self,
+      pcoll_name: str,
+      pcoll: beam.pvalue.PValue,
+      max_n: int,
+      max_duration_secs: float) -> Union[None, ElementStream]:
+    # noqa: F821
 
     """Reads an ElementStream of a computed PCollection.
 
