@@ -176,6 +176,7 @@ class SdkHarness(object):
       # that should be reported to the runner when proocessing the first bundle.
       deferred_exception=None,  # type: Optional[Exception]
       runner_capabilities=frozenset(),  # type: FrozenSet[str]
+      element_processing_timeout_minutes=None,  # type: Optional[int]
   ):
     # type: (...) -> None
     self._alive = True
@@ -207,6 +208,8 @@ class SdkHarness(object):
     self._profiler_factory = profiler_factory
     self.data_sampler = data_sampler
     self.runner_capabilities = runner_capabilities
+    self._element_processing_timeout_minutes = (
+        element_processing_timeout_minutes)
 
     def default_factory(id):
       # type: (str) -> beam_fn_api_pb2.ProcessBundleDescriptor
@@ -223,21 +226,21 @@ class SdkHarness(object):
         fns=self._fns,
         data_sampler=self.data_sampler,
     )
-
+    self._status_handler = None  # type: Optional[FnApiWorkerStatusHandler]
     if status_address:
       try:
         self._status_handler = FnApiWorkerStatusHandler(
             status_address,
             self._bundle_processor_cache,
             self._state_cache,
-            enable_heap_dump)  # type: Optional[FnApiWorkerStatusHandler]
+            enable_heap_dump,
+            element_processing_timeout_minutes=self.
+            _element_processing_timeout_minutes)
       except Exception:
         traceback_string = traceback.format_exc()
         _LOGGER.warning(
             'Error creating worker status request handler, '
             'skipping status report. Trace back: %s' % traceback_string)
-    else:
-      self._status_handler = None
 
     # TODO(BEAM-8998) use common
     # thread_pool_executor.shared_unbounded_instance() to process bundle
@@ -448,7 +451,7 @@ class BundleProcessorCache(object):
     self.known_not_running_instruction_ids = collections.OrderedDict(
     )  # type: collections.OrderedDict[str, bool]
     self.failed_instruction_ids = collections.OrderedDict(
-    )  # type: collections.OrderedDict[str, bool]
+    )  # type: collections.OrderedDict[str, Exception]
     self.active_bundle_processors = {
     }  # type: Dict[str, Tuple[str, bundle_processor.BundleProcessor]]
     self.cached_bundle_processors = collections.defaultdict(
@@ -537,9 +540,11 @@ class BundleProcessorCache(object):
     """
     with self._lock:
       if instruction_id in self.failed_instruction_ids:
+        e = self.failed_instruction_ids[instruction_id]
         raise RuntimeError(
             'Bundle processing associated with %s has failed. '
-            'Check prior failing response for details.' % instruction_id)
+            'Check prior failing response and attached exception for details.' %
+            instruction_id) from e
       processor = self.active_bundle_processors.get(
           instruction_id, (None, None))[-1]
       if processor:
@@ -548,14 +553,14 @@ class BundleProcessorCache(object):
         return None
       raise RuntimeError('Unknown process bundle id %s.' % instruction_id)
 
-  def discard(self, instruction_id):
-    # type: (str) -> None
+  def discard(self, instruction_id, exception):
+    # type: (str, Exception) -> None
 
     """
     Marks the instruction id as failed shutting down the ``BundleProcessor``.
     """
     with self._lock:
-      self.failed_instruction_ids[instruction_id] = True
+      self.failed_instruction_ids[instruction_id] = exception
       while len(self.failed_instruction_ids) > MAX_FAILED_INSTRUCTIONS:
         self.failed_instruction_ids.popitem(last=False)
       processor = self.active_bundle_processors[instruction_id][1]
@@ -599,7 +604,7 @@ class BundleProcessorCache(object):
       self.periodic_shutdown = None
 
     for instruction_id in list(self.active_bundle_processors.keys()):
-      self.discard(instruction_id)
+      self.discard(instruction_id, RuntimeError('Shutdown invoked'))
     for cached_bundle_processors in self.cached_bundle_processors.values():
       BundleProcessorCache._shutdown_cached_bundle_processors(
           cached_bundle_processors)
@@ -708,9 +713,9 @@ class SdkWorker(object):
       if not requests_finalization:
         self.bundle_processor_cache.release(instruction_id)
       return response
-    except:  # pylint: disable=bare-except
+    except Exception as e:  # pylint: disable=bare-except
       # Don't re-use bundle processors on failure.
-      self.bundle_processor_cache.discard(instruction_id)
+      self.bundle_processor_cache.discard(instruction_id, e)
       raise
 
   def process_bundle_split(
@@ -779,8 +784,8 @@ class SdkWorker(object):
         self.bundle_processor_cache.release(request.instruction_id)
         return beam_fn_api_pb2.InstructionResponse(
             instruction_id=instruction_id, finalize_bundle=finalize_response)
-      except:
-        self.bundle_processor_cache.discard(request.instruction_id)
+      except Exception as e:
+        self.bundle_processor_cache.discard(request.instruction_id, e)
         raise
     # We can reach this state if there was an erroneous request to finalize
     # the bundle while it is being initialized or has already been finalized
